@@ -1,5 +1,5 @@
 /*
-Copyright 2015 Esri
+Copyright 2015 - 2022 Esri
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
@@ -39,12 +39,12 @@ public:
   BitStuffer()  {}
   virtual ~BitStuffer()  {}
 
-  bool read(Byte** ppByte, std::vector<unsigned int>& dataVec) const;
+  bool read(const Byte** ppByte, std::vector<unsigned int>& dataVec) const;
 
 protected:
   mutable std::vector<unsigned int>  m_tmpBitStuffVec;
 
-  static bool readUInt(Byte** ppByte, unsigned int& k, int numBytes);    // numBytes = 1, 2, or 4
+  static bool readUInt(const Byte** ppByte, unsigned int& k, int numBytes);    // numBytes = 1, 2, or 4
   static unsigned int numTailBytesNotNeeded(unsigned int numElem, int numBits);
 };
 
